import typing

from that_depends.providers.base import AbstractProvider


T = typing.TypeVar("T")
P = typing.ParamSpec("P")


class Factory(AbstractProvider[T]):
    def __init__(self, factory: type[T] | typing.Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> None:
        self._factory = factory
        self._args = args
        self._kwargs = kwargs
        self._override = None

    async def async_resolve(self) -> T:
        if self._override:
            return typing.cast(T, self._override)

        return self._factory(
            *[await x.async_resolve() if isinstance(x, AbstractProvider) else x for x in self._args],
            **{k: await v.async_resolve() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()},
        )

    def sync_resolve(self) -> T:
        if self._override:
            return typing.cast(T, self._override)

        return self._factory(
            *[x.sync_resolve() if isinstance(x, AbstractProvider) else x for x in self._args],
            **{k: v.sync_resolve() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()},
        )


class AsyncFactory(AbstractProvider[T]):
    def __init__(self, factory: typing.Callable[P, typing.Awaitable[T]], *args: P.args, **kwargs: P.kwargs) -> None:
        self._factory = factory
        self._args = args
        self._kwargs = kwargs
        self._override = None

    async def async_resolve(self) -> T:
        if self._override:
            return typing.cast(T, self._override)

        return await self._factory(
            *[await x.async_resolve() if isinstance(x, AbstractProvider) else x for x in self._args],
            **{k: await v.async_resolve() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()},
        )

    def sync_resolve(self) -> T:
        msg = "AsyncFactory cannot be resolved synchronously"
        raise RuntimeError(msg)



import typing

from that_depends.providers.base import AbstractProvider


T = typing.TypeVar("T")


class List(AbstractProvider[list[T]]):
    def __init__(self, *providers: AbstractProvider[T]) -> None:
        self._providers = providers

    async def async_resolve(self) -> list[T]:
        return [await x.async_resolve() for x in self._providers]

    def sync_resolve(self) -> list[T]:
        return [x.sync_resolve() for x in self._providers]

    async def __call__(self) -> list[T]:
        return await self.async_resolve()



from that_depends.providers.base import AbstractProvider, AbstractResource
from that_depends.providers.collections import List
from that_depends.providers.context_resources import (
    AsyncContextResource,
    ContextResource,
    DIContextMiddleware,
    container_context,
)
from that_depends.providers.factories import AsyncFactory, Factory
from that_depends.providers.resources import AsyncResource, Resource
from that_depends.providers.singleton import Singleton


__all__ = [
    "AbstractProvider",
    "AbstractResource",
    "AsyncContextResource",
    "AsyncFactory",
    "AsyncResource",
    "ContextResource",
    "DIContextMiddleware",
    "Factory",
    "List",
    "Resource",
    "Singleton",
    "container_context",
]
